from langchain_core.tools import BaseTool
from langchain_core.tools import render_text_description
from langchain_core.runnables import RunnableSerializable
from langchain.prompts.chat import (
    ChatPromptTemplate,
    SystemMessagePromptTemplate,
    HumanMessagePromptTemplate,
)
from typing import List, Optional
from langchain_core.output_parsers import JsonOutputParser
import json
import re
from .parser import ReactStep, ReactKeywords


# React tool caller
class ReactToolCaller(RunnableSerializable[List[ReactStep], ReactStep]):
    tools: List[BaseTool] = []
    keywords: Optional[ReactKeywords] = ReactKeywords()
    refine_model: Optional[RunnableSerializable] = Optional
    refine_template: Optional[str] = ""

    def invoke(self, action: ReactStep, action_input: ReactStep) -> ReactStep:
        # Skip input with no actions and invalid input
        if (
            not action
            or not action_input
            or not action.type == "action"
            or not action_input.type == "action_input"
        ):
            raise Exception("Invalid tool calling input!", input)
        # Define arguments and tool
        tool_names = re.sub(r'[^A-Za-z_]', ' ', action.content).split()
        tool_name = tool_names[0] if tool_names else "undefined"
        tool_args = action_input.content
        tools = [x for x in self.tools if x.name == tool_name]
        if not tools:
            return ReactStep(type="observation", content="Error! Invalid tool name.")
        tool = tools[0]
        parser = JsonOutputParser(pydantic_object=tool.args)
        # Try parsing it directly
        try:
            tool_args = tool_args.replace("```json", "").replace("```", "")
            tool_args = parser.invoke(tool_args) if tool_args else tool_args
        # Use LLM to refine arguments
        except:
            if not self.refine_model:
                return ReactStep(
                    type="observation", content="Error! Failed parsing arguments."
                )
            # Prepare chain
            prompt = ChatPromptTemplate(
                messages=[
                    SystemMessagePromptTemplate.from_template(
                        self.refine_template, template_format="jinja2"
                    ),
                    HumanMessagePromptTemplate.from_template("{input}"),
                ],
                input_variables=["input", "tool_description", "output_format"],
            )
            chain = prompt | self.refine_model | parser
            try:
                tool_args = chain.invoke(
                    {
                        "input": tool_args,
                        "tool_description": render_text_description([tool]),
                        "output_format": json.dumps(
                            {key: value["type"] for key, value in tool.args.items()}
                        ),
                    },
                )
            except:
                return ReactStep(
                    type="observation", content="Error! Failed parsing arguments."
                )
        # Invoke tool
        try:
            result = tool.invoke(tool_args)
        except:
            return ReactStep(type="observation", content="Error! Tool call failed.")
        # Format result
        if isinstance(result, dict):
            result = json.dumps(result)
        return ReactStep(type="observation", content=result)
