from langchain_core.runnables import RunnableSerializable
from langchain.prompts.chat import (
    ChatPromptTemplate,
    SystemMessagePromptTemplate,
    HumanMessagePromptTemplate,
    AIMessagePromptTemplate,
)
from langchain_core.messages import AIMessage
from langchain_core.tools import BaseTool, render_text_description
from typing import List, Optional, Union, Iterator, Literal, Tuple
import importlib.resources
from .parser import ReactStep, ReactKeywords, ReactParser
from .caller import ReactToolCaller
import os

os.environ["LANGSMITH_TRACING"] = "true"
os.environ["LANGSMITH_API_KEY"] = "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"

# Default templates
DEFAULT_REACT_TEMPLATE = importlib.resources.read_text(
    "lcrai.templates", "react.jinja2", encoding="utf-8"
)

DEFAULT_REFINE_TEMPLATE = importlib.resources.read_text(
    "lcrai.templates", "refine.jinja2", encoding="utf-8"
)


# Improved react agent
class ReactAgentImproved(RunnableSerializable[str, List[ReactStep]]):
    model: RunnableSerializable
    tools: List[BaseTool] = []
    max_loops: int = 2
    system_prompt: Optional[Union[str, List[str]]] = None
    output_format: Optional[Literal["steps", "message"]] = "message"
    react_keywords: Optional[ReactKeywords] = ReactKeywords()
    react_template: Optional[str] = DEFAULT_REACT_TEMPLATE
    refine_tool_model: Optional[RunnableSerializable] = None
    refine_tool_template: Optional[str] = DEFAULT_REFINE_TEMPLATE

    # Return whole response
    def invoke(
        self, input: str, history: List[Tuple[str]] = []
    ) -> Union[List[ReactStep], AIMessage]:
        for chunk in self.stream(input, history):
            if not isinstance(chunk, str):
                return chunk

    # Stream for a single turn
    def stream_once(
        self,
        input: str,
        scratchpad: str,
        final: bool = False,
        history: List[Tuple[str]] = [],
    ) -> Iterator[Union[str, List[ReactStep]]]:
        # Define stop words
        stop_words = [
            f"\n{self.react_keywords.observation}:",
            f"\n{self.react_keywords.question}",
        ]
        if final:
            stop_words.extend(
                [
                    f"\n{self.react_keywords.action}:",
                    f"\n{self.react_keywords.thought}:",
                ]
            )
        self.model.__setattr__("stop", stop_words)
        # Define prefix
        prefix = f"{self.react_keywords.thought}: "
        if final:
            prefix = f"{prefix}I now know the final answer. {self.react_keywords.thought}: "
        yield prefix
        scratchpad = f"{scratchpad}\n{prefix}"
        # Define chain
        history = [
            (
                (
                    self.react_keywords.question
                    if role == "human"
                    else self.react_keywords.final_answer
                ),
                content,
            )
            for role, content in history
        ]
        prompt = ChatPromptTemplate(
            messages=[
                HumanMessagePromptTemplate.from_template(
                    DEFAULT_REACT_TEMPLATE, template_format="jinja2"
                ),
                #HumanMessagePromptTemplate.from_template("{input}"),
                AIMessagePromptTemplate.from_template("Question: {input}\n{scratchpad}"),
            ],
            input_variables=["input", "scratchpad"],
            partial_variables={
                "system_prompt": self.system_prompt if isinstance(self.system_prompt, str) else "\n\n".join(self.system_prompt),
                "history": history,
                "tools": render_text_description(self.tools) if self.tools else None,
                "tool_names": [x.name for x in self.tools] if self.tools else None,
            },
        )
        parser = ReactParser(keywords=self.react_keywords)
        generation_chain = prompt | self.model
        # Generate response
        resp = None
        for chunk in generation_chain.stream(
            {"input": input, "scratchpad": scratchpad},
            stream_usage=True,
        ):
            yield chunk.content
            if not resp:
                resp = chunk
            else:
                resp += chunk
        # Add prefix
        resp.content = f"\n{prefix}{resp.content}"
        # Parse output
        steps: List[ReactStep] = parser.invoke(resp)
        yield steps

    # ReAct loop
    def stream(
        self, input: str, history: List[Tuple[str]] = []
    ) -> Iterator[Union[str, List[ReactStep], AIMessage]]:
        # Define parser and tool caller
        parser = ReactParser(tools=self.tools, keywords=self.react_keywords)
        caller = ReactToolCaller(
            tools=self.tools,
            keywords=self.react_keywords,
            refine_model=(
                self.refine_tool_model if self.refine_tool_model else self.model
            ),
            refine_template=self.refine_tool_template,
        )
        # Loop
        steps: List[ReactStep] = []
        n_loop: int = 0
        while n_loop < self.max_loops:
            # Generate message
            is_final = n_loop + 1 >= self.max_loops
            for chunk in self.stream_once(input, parser.dump(steps), is_final, history):
                if isinstance(chunk, list):
                    steps = parser.merge(steps, chunk)
                else:
                    yield chunk
            # Fail case
            if not steps:
                raise Exception("No response from API")
            # Call tools
            if steps[-1].type == "action":
                new_step = caller.invoke(
                    steps[-1], ReactStep(type="action_input", content="")
                )
                yield f"\n{new_step.dump(self.react_keywords)}\n"
                steps.append(new_step)
            elif steps[-1].type == "action_input" and steps[-2].type == "action":
                new_step = caller.invoke(steps[-2], steps[-1])
                yield f"\n{new_step.dump(self.react_keywords)}\n"
                steps.append(new_step)
            # Finish
            elif steps[-1].type == "final_answer":
                break
            n_loop += 1
        # Output data in the right format
        if self.output_format == "steps":
            steps = [ReactStep(type="question", content=input)] + steps
            yield steps
        else:
            usage = steps[-1].usage if steps[-1].usage else None
            yield AIMessage(steps[-1].content, usage_metadata=usage)
