from langchain_core.messages import AIMessage
from langchain_core.runnables import RunnableSerializable
from langchain_core.tools import BaseTool
from pydantic import BaseModel, field_validator
import re
from typing import List, Optional

# Define constants
USAGE_METADATA_KEYS = ["input_tokens", "output_tokens", "total_tokens"]


# Define react keyword collection
class ReactKeywords(BaseModel):
    question: str = "Question"
    action: str = "Action"
    action_input: str = "Action Input"
    final_answer: str = "Final Answer"
    thought: str = "Thought"
    observation: str = "Observation"
    undefined: str = "Undefined"


# React step structure
class ReactStep(BaseModel):
    type: str
    content: str
    usage: Optional[dict[str, int]] = None

    # Clean usage metadata
    @field_validator("usage", mode="before")
    def clean_usage(cls, usage: Optional[dict]) -> Optional[dict[str, int]]:
        if not usage:
            return None
        return {key: int(usage[key]) for key in USAGE_METADATA_KEYS if key in usage}

    # Strip content
    @field_validator("content", mode="before")
    def preprocess_content(cls, content: str) -> str:
        return content.strip()

    # Parse from plain text
    @classmethod
    def from_text(
        cls, text: str, keywords: Optional[dict] = None, usage: Optional[dict] = None
    ):
        parts = text.split(":")
        if len(parts) < 2:
            return cls(type="undefined", content=parts[0] if parts else "", usage=usage)
        step_type = keywords[parts[0].lower()]
        content = ":".join(parts[1:])
        for key in keywords.values():
            if content.lower().startswith((key+":").lower()):
                content = content[len(key+":"):]
        return cls(type=step_type, content=content, usage=usage)

    # Dump to plain text
    def dump(self, keywords: ReactKeywords = ReactKeywords()):
        return f"{keywords.__getattribute__(self.type)}: {self.content}"


# React prompt parser
class ReactParser(RunnableSerializable[AIMessage, List[ReactStep]]):
    keywords: Optional[ReactKeywords] = ReactKeywords()

    # Invoke method
    def invoke(self, input: AIMessage) -> List[ReactStep]:
        # Ignore empty
        if not input.content or input.content == "":
            return []
        # Create reversed keyword lookup dict
        keywords = {
            value.lower(): key.lower()
            for key, value in self.keywords.model_dump().items()
        }
        # Define start indices for each possible step
        lower_content = input.content.lower()
        starts: List[int] = []
        for keyword in keywords.keys():
            starts += [x.start() for x in re.finditer(f"\n{keyword}:", lower_content)]
        starts.sort()
        # Parse steps
        steps: List[ReactStep] = []
        for i in range(len(starts)):
            end = starts[i + 1] if i + 1 < len(starts) else len(input.content)
            step_text = input.content[starts[i] : end].strip()
            usage = None
            if i == len(starts) - 1 and input.usage_metadata:
                usage = input.usage_metadata
            steps.append(ReactStep.from_text(step_text, keywords, usage))
        return steps

    # Dump as string
    def dump(self, input: List[ReactStep]) -> str:
        return "\n".join([x.dump(self.keywords) for x in input])

    # Merge two lists preserving usage data
    def merge(
        self, input1: List[ReactStep], input2: List[ReactStep]
    ) -> List[ReactStep]:
        # Add steps
        input1.extend(input2)
        # Find usage details
        usage = None
        for step in input1:
            if not step.usage:
                continue
            if not usage:
                usage = step.usage
            else:
                for key in USAGE_METADATA_KEYS:
                    usage[key] += step.usage[key] if key in step.usage else 0
            step.usage = None
        # Move usage details to the last step
        if usage and input1:
            input1[-1].usage = usage
        return input1
